/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn
 */

#pragma once

namespace cusp
{
namespace detail
{
namespace device
{
static __inline__ __device__ float atomicAdd(float *addr, float val)
{
    float old=*addr, assumed;
    
    do {
        assumed = old;
        old = int_as_float( atomicCAS((int*)addr,
                                        float_as_int(assumed),
                                        float_as_int(val+assumed)));
    } while( float_as_int(assumed)!=float_as_int(old) );
    //
    // NOTE: Comparing as ints rather than floats is mandatory.
    // For floats, the test NaN==NaN would always fall, leading to an
    // infinite loop.

    return old;
}

static __inline__ __device__ double atomicAdd(double *addr, double val)
{
    double old=*addr, assumed;
    
    do {
        assumed = old;
        old = __longlong_as_double( atomicCAS((unsigned long long int*)addr,
                                        __double_as_longlong(assumed),
                                        __double_as_longlong(val+assumed)));
    } while( __double_as_longlong(assumed)!=__double_as_longlong(old) );

    return old;
}

}
template <typename IndexType, typename ValueType>
void subscript(const IndexType n, 
			   const IndexType *indices, 
			   const ValueType *x, 
					 ValueType *y,
					 cusp::device_memory);
					 
template <typename IndexType, typename ValueType>
void subscript(const IndexType n, 
			   const IndexType *indices, 
			   const ValueType *x, 
					 ValueType *y,
					 cusp::host_memory);
					 
template <typename IndexArray>
void indices_to_nnz(const IndexArray& indices, IndexArray& nnz);

template <typename IndexArray>
void nnz_to_offsets(const IndexArray& nnz, IndexArray& offsets);					 
					 
template <typename MatrixType>
void extract_triangle(const MatrixType& A, MatrixType& A_lo, MatrixType& A_up);

template <typename Matrix>
void sort_by_row_and_column(Matrix& mat);

// convert Modified Sparse Row (MSR) format to CSR format
template <typename IndexType, typename ValueType, typename MatrixType>
void msr_to_csr(const IndexType num_rows,
			 const IndexType num_cols,
			 const IndexType num_entries, 
			 const IndexType *ju,
			 const IndexType *jau,			 
			 const ValueType *au,
			 bool  to_mkl,
			 MatrixType& A_csr);

template <typename IndexType, typename ValueType, typename Matrix>
void setup_mkl_matrix(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& dst, Matrix& src);

} // end namespace detail
} // end namespace cusp

#include <cusp/detail/utils_.inl>

