/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file permutation.h
 *  \brief permute array or matrix
 */

namespace cusp {

// permutate by scatter
// pass pt as backward: from objective to original
// pass p as forward: from original to objective
template< typename Array1, typename Array2 >
void permute_array(const Array1 &p, const Array2 &x, Array2 &y);

// get the inverse permutation
template< typename Array1, typename Array2 >
void get_inverse_permutation(const Array1 &p, Array2 &pt);

template< typename Array1d, typename Matrix >
void build_permutation_matrix(const Array1d &p, Matrix &A);

template< typename Matrix, typename Array1d >
void compute_ptAqt(const Matrix &A, const Array1d &p, const Array1d &q, Matrix &ptAqt);

template< typename Matrix, typename Array1d >
void compute_ptAp(const Matrix &A, const Array1d &p, Matrix &ptAp);

template< typename Matrix1, typename Matrix2 >
void compute_ptAp(const Matrix1 &A, const Matrix2 &pt_matrix, const Matrix2 &p_matrix, Matrix1 &ptAp);

} // namespace cusp

#include <cusp/detail/permutation.inl>
