/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

#pragma once

#include <cusp/detail/config.h>

#include <cusp/csr_matrix.h>
#include <cusp/coo_matrix.h>

#include <cusp/precond/smooth.h>

namespace cusp
{
namespace precond
{
	// the easiest way to use Reconstruction function. It is convient for different format and memoryspace used here.
	// for general cases, look forward to permuation.inl. 
	// for the template usage: it is better to use MatrixType, then get IndexType from MatrixType.
	// but if the parameters need IndexType, it should be presented in templates
	template <typename IndexType, typename ValueType>
	void bilinear_prolongator(cusp::coo_matrix<IndexType,ValueType,cusp::device_memory>& P,
				IndexType& mx_c, IndexType& my_c,
				const IndexType mx, const IndexType my,
				const IndexType ratioi, const IndexType ratioj, const IndexType dof);
				
	template <typename IndexType, typename ValueType>
	void bilinear_prolongator(cusp::csr_matrix<IndexType,ValueType,cusp::host_memory>& P,
				IndexType& mx_c, IndexType& my_c,
				const IndexType mx, const IndexType my,
				const IndexType ratioi, const IndexType ratioj, const IndexType dof);

} // end namespace precond
} // end namespace cusp

#include <cusp/precond/detail/bilinear.inl>
