/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file hierarchical_as.inl
 *  \brief Inline file for hierarchical_as.h
 */

namespace cusp
{
namespace relaxation
{
template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename MatrixType, typename VectorType>	
hierarchical_as<IndexType, ValueType, MemorySpace>::hierarchical_as(
	bool same_mesh, 
	bool same_structure, 	
	const MatrixType& A, 
	const VectorType& domains, 
	const IndexType overlap) :
    local_solver(same_mesh, same_structure, A, domains, overlap)
{
	initialized = true;		
}

template <typename IndexType, typename ValueType, typename MemorySpace>
template <typename MatrixType, typename VectorType>	
void hierarchical_as<IndexType, ValueType, MemorySpace>::setup(
	const MatrixType& A, 
	const VectorType& domains, 
	bool type, 
	const IndexType overlap, 
	const IndexType sweeps,
	const IndexType iter, 
	const ValueType jacobi_tol)
{
	if (!initialized) {
		 printf("hierarchical_as has not yet been created!\n");
		 exit(1);		
	}
	
	local_solver.setup(A, domains, type?(cusp::precond::asm_type)1:(cusp::precond::asm_type)0, overlap, sweeps, iter, jacobi_tol);
}

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void hierarchical_as<IndexType,ValueType,MemorySpace>
    ::operator()(const MatrixType &A, const VectorType1& b, VectorType2& x)
{
  local_solver.relax(b, x);
}

template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void hierarchical_as<IndexType,ValueType,MemorySpace>
    ::presmooth(const MatrixType &A, const VectorType1& b, VectorType2& x)
{
  // call the asm preconditioner without matrix
  local_solver.presmooth(A, b, x);
}



template <typename IndexType, typename ValueType, typename MemorySpace>
template<typename MatrixType, typename VectorType1, typename VectorType2>
    void hierarchical_as<IndexType,ValueType,MemorySpace>
    ::postsmooth(const MatrixType &A, const VectorType1& b, VectorType2& x)
{
  local_solver.postsmooth(A, b, x);
}

} // end namespace relaxation
} // end namespace cusp

