/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file hierarchical_as.h
 *  \brief Jacobi relaxation.
 */

#pragma once

#include <cusp/precond/asm.h>

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
class hierarchical_as
{
    cusp::precond::additive_schwarz<IndexType, ValueType, MemorySpace> local_solver; 

public:

	bool initialized;
    hierarchical_as();

	template <typename MatrixType, typename VectorType>	
    hierarchical_as(
	bool same_mesh, 
	bool same_structure, 
	const MatrixType& A, 
	const VectorType& domains, 
	const IndexType overlap=1);	 

	template <typename MatrixType, typename VectorType>	
    void setup(
	const MatrixType& A, 
	const VectorType& domains, 
	bool type=true, 
	const IndexType overlap=1, 
	const IndexType sweeps=1,
	const IndexType iter=1000, 
	const ValueType jacobi_tol=1e-12);
	
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void operator()(const MatrixType &A, const VectorType1& b, VectorType2& x);
    
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType &A, const VectorType1& b, VectorType2& x);
    
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType &A, const VectorType1& b, VectorType2& x);
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/hierarchical_as.inl>
