/*
 *  Copyright 2008-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file hierarchical_jacobi.h
 *  \brief Jacobi relaxation.
 */

#pragma once

#include <cusp/precond/additive_schwarz.h>

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
class hierarchical_jacobi
{
    cusp::precond::additive_schwarz<IndexType, ValueType, MemorySpace> local_solver; 
    bool _initialized;

public:
    hierarchical_jacobi();

    hierarchical_jacobi(
      const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A, 
      const cusp::array1d<IndexType,MemorySpace> &domains, 
      int inner_iterations, 
      int outer_iterations,
      ValueType omega);

   void initialize(
      const cusp::csr_matrix<IndexType,ValueType,MemorySpace>& A, 
      const cusp::array1d<IndexType,MemorySpace> &domains, 
      int inner_iterations, 
      int outer_iterations,
      ValueType omega);

    
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void operator()(const MatrixType &A, const VectorType1& b, VectorType2& x);
    
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType &A, const VectorType1& b, VectorType2& x);
    
    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType &A, const VectorType1& b, VectorType2& x);
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/hierarchical_jacobi.inl>

