/*
 *  Copyright 2010 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*! \file line_relaxation.h
 *  \brief line relaxation.
 */

#pragma once

#include <cusp/detail/config.h>

#include <cusp/linear_operator.h>
#include <cusp/graph/path_cover.h>

#include <cusp/relaxation/jacobi.h>
#include <cusp/hyb_matrix.h>
#include <cusp/csr_matrix.h>

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
class line
{

public:
    //! constructor
    line(const csr_matrix<IndexType, ValueType, MemorySpace>& A, const cusp::graph::path_cover<IndexType, ValueType, MemorySpace>& _cover, int _sweep_count=3, ValueType _alpha=0.0);

    //! initialization procedure
    void init(const csr_matrix<IndexType, ValueType, MemorySpace>& A);

    //! linear_operator which applies the line relaxation
    //! \param bb current RHS vector
    //! \param xx vector which shall be smoothed
    template <typename VectorType1, typename VectorType2>
    void operator()(const VectorType1& bb, VectorType2& xx) const;
        
    const cusp::graph::path_cover<IndexType,ValueType,MemorySpace> &cover;
    thrust::device_vector<ValueType> a,b,c;

private:
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A_mat;   // copy of A
    cusp::hyb_matrix<IndexType, ValueType, MemorySpace> A_rest; // matrix that builds the rhs
    int sweep_count;

    ValueType alpha;
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/line_relaxation.inl>

