/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn 
 */

/*! \file mkl_ilu0.h
 *  \brief relaxed by MKL dcsrilu0.
 *  \only double precision is supported
 */

#pragma once

#include <cusp/detail/config.h>
#include <cusp/linear_operator.h>
#include <cusp/csr_matrix.h>
#include <cusparse.h>
#include <mkl_blas.h>
#include <mkl_cblas.h>
#include <mkl_spblas.h>
#include <mkl_rci.h>

#if defined(MKL_ILP64)
#define MKL_INT long long
#else
#define MKL_INT int
#endif

namespace cusp
{
namespace relaxation
{

template <typename IndexType, typename ValueType, typename MemorySpace>
class mkl_ilu0 : public cusp::linear_operator<ValueType, MemorySpace, IndexType>
{
	// MKL context
	MKL_INT ivar, *ipar, ierr;
	char cvar,cvar1,cvar2;	
	ValueType *dpar;
	cusp::array1d<ValueType,cusp::host_memory> val;
	// if use_mkl_sv==true, the container for matrix on host
	cusp::csr_matrix<IndexType,ValueType,cusp::host_memory> A_fact_mkl;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_r;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_y;
	cusp::array1d<ValueType,cusp::host_memory> temp_mkl_x;
		
	// cuSAPRSE context	
    cusparseHandle_t handle;
    cusparseStatus_t status;
	cusparseMatDescr_t descrM_lo;
	cusparseMatDescr_t descrM_up;
	cusparseSolveAnalysisInfo_t Analy_lo;
	cusparseSolveAnalysisInfo_t Analy_up;	
	// lower and upper part of factorization matrix, used in the triangular solver of cuSPARSE 
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_lo;
	cusp::csr_matrix<IndexType,ValueType,MemorySpace> A_up;
	cusp::array1d<ValueType,MemorySpace> temp_y;
	
public:

	bool initialized;
	bool same_structure;		
	// use MKL routine mkl_dcsrtrsv to solve the triangular systems
	bool use_mkl_sv;
	IndexType sweeps;
	
    mkl_ilu0();
	
    template <typename MatrixType>
    mkl_ilu0(bool same_structure, const MatrixType& A, cusparseHandle_t& handle, bool use_mkl_sv=false);

    ~mkl_ilu0();
	
    template <typename MatrixType>
    void setup(const MatrixType& A, const IndexType sweeps=1);
    
    // ignores initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
   
    // smooths initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename Array1, typename Array2>
    void operator()(const Array1& b, Array2& x);
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/mkl_ilu0.inl>

