/*
 *  Author: Li Luo 
 *  Email:  li.luo@siat.ac.cn  
*/

/*! \file polynomial_sweeps.h
 *  \brief polynomial relaxation with Richardson sweeps.
 */

#pragma once

#include <cusp/detail/config.h>

#include <cusp/linear_operator.h>

namespace cusp
{
namespace relaxation
{

template <typename ValueType, typename MemorySpace>
class polynomial 
{
    cusp::array1d<ValueType, cusp::host_memory> default_coefficients;
	int sweeps;
	
public:

	bool initialized;
	
    polynomial();

    template <typename VectorType>	
    void setup(const VectorType& coefficients, const int sweeps=1);
	
    // ignores initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void presmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);
   
    // smooths initial x
    template<typename MatrixType, typename VectorType1, typename VectorType2>
    void postsmooth(const MatrixType& A, const VectorType1& b, VectorType2& x);

    template <typename MatrixType, typename VectorType1, typename VectorType2>
    void operator()(const MatrixType& A, const VectorType1& b, VectorType2& x) const;

    template <typename MatrixType, typename VectorType1, typename VectorType2, typename VectorType3>
    void operator()(const MatrixType& A, const VectorType1& b, VectorType2& x, VectorType3& coeffients) const;
};

} // end namespace relaxation
} // end namespace cusp

#include <cusp/relaxation/detail/polynomial_sweeps.inl>

