#include <cusp/print.h>
#include <cusp/relaxation/ilupp.h>
#include <cusp/io/matrix_market.h>
#include <cusp/blas.h>
#include <cusp/krylov/cg.h>
#include <cusp/krylov/gmres.h>
#include <cusp/gallery/poisson.h>
#include <cusp/csr_matrix.h>
#include <cusp/precond/diagonal.h>
#include <iostream>
#include <performance/timer.h>
#include <cmath>

#define PI 3.1415926

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

int main(void)
{
    typedef int                 IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory MemorySpace;

    // create an empty sparse matrix structure
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A;

        bool matrix_market=false;
		int mm=-1;
		while(mm<0)
		{
			printf("Please input mm: 0: poisson2d,  1: UF collection \n");
			scanf("%d", &mm);
		}
		if(mm) {
		matrix_market=true;
		} else {
		matrix_market=false;		
		}
		
	std::string mtx="5pt_10x10.mtx";
	int m = 0, n = 0;
	
	if (matrix_market)
	{
    // load a matrix stored in MatrixMarket format
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);	
	} else {
	while(m < 2 || n < 2)
    {
    	printf("Please input grid size (m, n): ");
    	scanf("%d %d", &m, &n);
    }	
    // create 2D Poisson problem
    cusp::gallery::poisson5pt(A, m, n);	
	}

	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);
		
#if 0
    // solve with diagonal preconditioner
    {
        std::cout << "\nSolving with diagonal preconditioner (M = D^-1)" << std::endl;
        
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-7);

        // setup preconditioner
        cusp::precond::diagonal<ValueType, MemorySpace> M(A);

        // solve			
		timer t1;	
		//cusp::krylov::gmres(A, x, b, restart, monitor, M);
		cusp::krylov::cg(A, x, b, monitor, M);
		std::cout << "\nsolve by jacobi-cg in " << t1.milliseconds_elapsed() << " ms " << std::endl;	
		
		// report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }
#endif	
		
    // solve with iluc preconditioner 

		int nk=0;
		while(nk < 1)
		{
			printf("\nPlease input test times: ");
			scanf("%d", &nk);
		}
	for(int k=0;k<nk;k++)
    // solve with multigrid algebraic multigrid preconditioner
    {
        std::cout << "\n ******Solving with ilupp preconditioner****** " << std::endl;

        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
		
		// solve with ilu preconditioner 
		int ilu_type = -1;
		while(ilu_type < 0 || ilu_type > 2)
		{
			printf("\nPlease input ilu_type: 0: iluc, 1: ilut, 2: multilevel_ilu ");
			scanf("%d", &ilu_type);
		}		
		int sweeps=0;
		while(sweeps < 1)
		{
			printf("\nPlease input smoothing sweeps: ");
			scanf("%d", &sweeps);
		}
		int fill_in=-1;
		while(fill_in < 0)
		{
			printf("\nPlease input fill_in: ");
			scanf("%d", &fill_in);
		}
		ValueType threshold=-0.0001;		
		while(threshold < 0.)
		{
			printf("\nPlease input threshold: ");
			scanf("%lf", &threshold);
		}	
		int prepro_type=-1;
		while(prepro_type < 0)
		{
			printf("\nPlease input prepro_type: ");
			scanf("%d", &prepro_type);
		}
		
		bool use_mkl_sv=false;
		int sv=-1;
		if (ilu_type == 0 || ilu_type == 1) {
		while(sv<0)
		{
			printf("Please input sv: 0: use cuSPARSE,  1: use MKL to solve triangular system\n");
			scanf("%d", &sv);
		}
		if(sv) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}
		}
		bool detail_param=false;
		int dp=-1;
		if (ilu_type == 2) {
		while(dp<0)
		{
			printf("\nPlease input detail_param: 0: disable detail parameters,  1: enable\n");
			scanf("%d", &dp);
		}
		if(dp) {
		detail_param=true;
		} else {
		detail_param=false;		
		}
		}
		

		int precond_no;
		int max_levels = 0;
		ValueType mem_factor=-0.0001;	
		ValueType min_pivot=-0.0001;	
		bool max_fillin_is_inf=false;	
		int max_fi_inf=-1;		
		bool begin_total_piv=false;	
		int begin_tp=-1;		
		
		if (ilu_type == 2) {
		printf("\nPlease input precond_no: ");
		scanf("%d", &precond_no);	
		while(mem_factor < 0.)
		{
			printf("\nPlease input mem_factor: ");
			scanf("%lf", &mem_factor);
		}	
		
		if (detail_param) {		
		while(max_levels < 1)
		{
			printf("\nPlease input max_levels(>0): ");
			scanf("%d", &max_levels);
		}				
		while(min_pivot < 0.)
		{
			printf("\nPlease input min_pivot: ");
			scanf("%lf", &min_pivot);
		}	
		while(max_fi_inf<0)
		{
			printf("\nPlease input max_fillin_is_inf: 0: max_fillin_is_inf is false,  1: max_fillin_is_inf is true\n");
			scanf("%d", &max_fi_inf);
		}
		if(max_fi_inf) {
		max_fillin_is_inf=true;
		} else {
		max_fillin_is_inf=false;		
		}	
		while(begin_tp<0)
		{
			printf("\nPlease input begin_total_piv: 0: begin_total_piv is false,  1: begin_total_piv is true\n");
			scanf("%d", &begin_tp);
		}
		if(begin_tp) {
		begin_total_piv=true;
		} else {
		begin_total_piv=false;		
		}		
		} else {
		max_levels = 100;
		mem_factor = 3.0;	
		min_pivot  = 1e-2;	
		max_fillin_is_inf=true;	
		begin_total_piv=true;	
		}	
		}
		
		
	    cusparseHandle_t handle = 0;
		/* Get handle to the CUSPARSE context */
		cusparseCreate(&handle);
		
        // setup preconditioner
		timer t1;
		cusp::relaxation::ilupp<int,ValueType,MemorySpace>* M = 
		new cusp::relaxation::ilupp<int,ValueType,MemorySpace>(true, A, handle, use_mkl_sv, (cusp::relaxation::ilupp_type)ilu_type,
		prepro_type,precond_no,max_levels,mem_factor,threshold,min_pivot,max_fillin_is_inf,fill_in,begin_total_piv,detail_param);

		M->setup(A,sweeps);	
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;
 
		int solver=-1;
		while(solver < 0)
		{
			printf("\nPlease input solver: 0: cg, 1: gmres ");
			scanf("%d", &solver);
		}
		int restart=-1;
		if (solver==1) {
		while(restart < 0)
		{
			printf("\nPlease input gmres restart: ");
			scanf("%d", &restart);
		}
		}	
		
        // set stopping criteria (iteration_limit = 2000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 2000, 1e-7);
		
        // solve
		timer t2;
		if (solver==0)
        cusp::krylov::cg(A, x, b, monitor, *M);
		else if (solver==1)
        cusp::krylov::gmres(A, x, b, restart, monitor, *M);		
        std::cout << "\nsolve by cg multi-level ilu in " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
		
		delete M;
		cusparseDestroy(handle);		
    }
	
    return 0;
}

