#include <cusp/print.h>
#include <cusp/relaxation/mkl_ilu0.h>
#include <cusp/io/matrix_market.h>
#include <cusp/blas.h>
#include <cusp/krylov/cg.h>
#include <cusp/gallery/poisson.h>
#include <cusp/csr_matrix.h>
#include <cusp/precond/diagonal.h>
#include <iostream>
#include <performance/timer.h>
#include <cmath>
#define PI 3.1415926

template <typename Monitor>
void report_status(Monitor& monitor)
{
    if (monitor.converged())
    {
        std::cout << "  Solver converged to " << monitor.tolerance1() << " tolerance";
        std::cout << " after " << monitor.iteration_count() << " iterations";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
    else
    {
        std::cout << "  Solver reached iteration limit " << monitor.iteration_limit() << " before converging";
        std::cout << " to " << monitor.tolerance1() << " tolerance ";
        std::cout << " (" << monitor.relative_residual_norm() << " final residual)" << std::endl;
    }
}

int main(void)
{
    typedef int                 IndexType;
    typedef double               ValueType;
    typedef cusp::device_memory MemorySpace;

    // create an empty sparse matrix structure
    cusp::csr_matrix<IndexType, ValueType, MemorySpace> A;

        bool matrix_market=false;
		int mm=-1;
		while(mm<0)
		{
			printf("Please input mm: 0: poisson2d,  1: UF collection \n");
			scanf("%d", &mm);
		}
		if(mm) {
		matrix_market=true;
		} else {
		matrix_market=false;		
		}
		
	std::string mtx="5pt_10x10.mtx";
	int m = 0, n = 0;
	
	if (matrix_market)
	{
    // load a matrix stored in MatrixMarket format
	printf("Please input matrix : ");
	std::cin >> mtx;
    cusp::io::read_matrix_market_file(A, mtx);	
	} else {
	while(m < 2 || n < 2)
    {
    	printf("Please input grid size (m, n): ");
    	scanf("%d %d", &m, &n);
    }	
    // create 2D Poisson problem
    cusp::gallery::poisson5pt(A, m, n);	
	}

	// allocate storage for solution (x) and right hand side (b)	
	cusp::array1d<ValueType, MemorySpace> x_solution(A.num_rows, 1);
    // initialize starting vector to random values in [0,1)
    //cusp::copy(cusp::detail::random_reals<ValueType>(A.num_rows), x_solution);
	ValueType *x_so_cpu;	
	cudaMallocHost((void **) &x_so_cpu, (A.num_rows)*sizeof(ValueType));
		
	for(int i=0;i<A.num_rows;i++)	
	{
		x_so_cpu[i]=std::sin(0.5*PI*i/A.num_rows);
	}
	ValueType *x_so_gpu = thrust::raw_pointer_cast(&x_solution[0]);
	cudaMemcpy(x_so_gpu, x_so_cpu, (A.num_rows)*sizeof(ValueType), cudaMemcpyHostToDevice);
	cudaFreeHost(x_so_cpu);
	
    cusp::array1d<ValueType, MemorySpace> b(A.num_rows);
	cusp::multiply(A, x_solution, b);
		
#if 0
    // solve with diagonal preconditioner
    {
        std::cout << "\nSolving with diagonal preconditioner (M = D^-1)" << std::endl;
        
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);

        // set stopping criteria (iteration_limit = 10000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 10000, 1e-7);

        // setup preconditioner
        cusp::precond::diagonal<ValueType, MemorySpace> M(A);

        // solve			
		timer t1;	
		//cusp::krylov::gmres(A, x, b, restart, monitor, M);
		cusp::krylov::cg(A, x, b, monitor, M);
		std::cout << "\nsolve by jacobi-cg in " << t1.milliseconds_elapsed() << " ms " << std::endl;	
		
		// report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
    }
#endif	
		
    // solve with ilu0 preconditioner 

		int nk=0;
		while(nk < 1)
		{
			printf("\nPlease input test times: ");
			scanf("%d", &nk);
		}
	for(int k=0;k<nk;k++)
    // solve with multigrid algebraic multigrid preconditioner
    {
        std::cout << "\n ******Solving with ilu0 preconditioner, PETSc sort****** " << std::endl;

        // allocate storage for solution (x) and right hand side (b)
        cusp::array1d<ValueType, MemorySpace> x(A.num_rows, 0);
		
		bool use_mkl_sv=false;
		int sv=-1;
		while(sv<0)
		{
			printf("Please input sv: 0: use cuSPARSE,  1: use MKL to solve triangular system\n");
			scanf("%d", &sv);
		}
		if(sv) {
		use_mkl_sv=true;
		} else {
		use_mkl_sv=false;		
		}
		int sweeps=0;
		while(sweeps < 1)
		{
			printf("\nPlease input smoothing sweeps: ");
			scanf("%d", &sweeps);
		}

	    cusparseHandle_t handle = 0;
		/* Get handle to the CUSPARSE context */
		cusparseCreate(&handle);
		
        // setup preconditioner
		timer t1;
		cusp::relaxation::mkl_ilu0<int,ValueType,MemorySpace>* M = 
		new cusp::relaxation::mkl_ilu0<int,ValueType,MemorySpace>(true, A, handle, use_mkl_sv);  
		M->setup(A,sweeps);	
        std::cout << "\nconstruct preconditioner in " << t1.milliseconds_elapsed() << " ms " << std::endl;

        // set stopping criteria (iteration_limit = 2000, relative_tolerance = 1e-7)
        cusp::verbose_monitor<ValueType> monitor(b, 2000, 1e-7);
        
        // solve
		timer t2;
        cusp::krylov::cg(A, x, b, monitor, *M);
        std::cout << "\nsolve by as in " << t2.milliseconds_elapsed() << " ms " << std::endl;
		
        // report status
        report_status(monitor);
		
		// check solution
        cusp::array1d<ValueType, MemorySpace> e(A.num_rows);
		blas::axpby(x_solution, x, e, ValueType(1), ValueType(-1));
		ValueType e_norm = cusp::blas::nrm2(e);
		printf("\n check solution: ||x*-x||: %f \n", e_norm );	
		
		cusp::array1d<ValueType, MemorySpace> ax(x.size());
		cusp::array1d<ValueType, MemorySpace> r(A.num_rows);
		cusp::multiply(A, x, ax);
		cusp::blas::axpby(ax, b, r, ValueType(1), ValueType(-1));
		ValueType r_norm = cusp::blas::nrm2(r);
		ValueType b_norm = cusp::blas::nrm2(b);		
		printf("\n ||b-Ax||/||b||: %f \n", r_norm / b_norm );
		
		delete M;
		cusparseDestroy(handle);		
    }
	
    return 0;
}

