/***************************************************************************
 *   Copyright (C) 2007 by Jan Mayer                                       *
 *   jan.mayer@mathematik.uni-karlsruhe.de                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



// this program uses only the ILU++ interface to allow easy use in other programs.
// it requires the iluplusplus library including the most important features of ILU++

// include the classes and functions defined in the library iluplusplus(version).a

#include "iluplusplus_interface.h"

typedef iluplusplus::Coeff_Field Coeff_Field;

int main(int argc, char *argv[]){
  try {
  // *************************************************************************** //
  // PART I: solving linear systems stored in Harwell-Boeing format.             //
  // *************************************************************************** //

    if(argc != 10){
        std::cout<<"main_solve_interface: arguments: <preprocessing number> <preconditioner number> <memory factor> <fill in> <-log10(threshold)> <matrix file> <output directory> <subdirectory output data> <matrix name>"<<std::endl;
        return 1;
    }
    iluplusplus::preprocessing_sequence L;
    iluplusplus::iluplusplus_precond_parameter param;

    switch ( atoi(argv[1]) ){
        case -1:  L.set_none(); break;
        case  0:  L.set_normalize(); break;
        case  1:  L.set_PQ(); break;
        case  2:  L.set_MAX_WEIGHTED_MATCHING_ORDERING(); break;
        case  3:  L.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
        case  4:  L.set_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        case  5:  L.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG_DD_MOV_COR_IM(); break;
        case  6:  L.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING(); break;
        case  7:  L.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
        case  8:  L.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        case  9:  L.set_SPARSE_FIRST_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG_DD_MOV_COR_IM(); break;
        case 10:  L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING(); break;
        case 11:  L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        #ifdef ILUPLUSPLUS_USES_METIS
        case 12:  L.set_METIS_NODE_ND_ORDERING(); break;
        case 13:  L.set_PQ_METIS_NODE_ND_ORDERING(); break;
        #endif
        case 14: L.set_MAX_WEIGHTED_MATCHING_ORDERING(); break;
        case 15: L.set_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
        case 16: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
        case 17: L.set_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
        case 18: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
        case 19: L.set_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
        case 20: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
        case 21: L.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
        case 22: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
        case 23: L.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
        case 24: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
        case 25: L.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
        case 26: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
        case 27: L.set_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
        case 28: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
        case 29: L.set_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        case 30: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        case 31: L.set_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
        case 32: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
        case 33: L.set_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
        case 34: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
        case 35: L.set_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
        case 36: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
        case 37: L.set_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
        case 38: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
        case 39: L.set_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;
        case 40: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;
        #ifdef ILUPLUSPLUS_USES_METIS
        case 41: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER(); break;
        case 42: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER(); break;
        case 43: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER_IM(); break;
        case 44: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER_IM(); break;
        case 45: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR(); break;
        case 46: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR(); break;
        case 47: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR_IM(); break;
        case 48: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR_IM(); break;
        case 49: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_DD_MOV_COR_IM(); break;
        case 50: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_DD_MOV_COR_IM(); break;
        case 51: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR(); break;
        case 52: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR(); break;
        case 53: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR_IM(); break;
        case 54: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR_IM(); break;
        case 55: L.set_MAX_WEIGHTED_MATCHING_ORDERING_METIS(); break;
        case 56: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_METIS(); break;
        case 57: L.set_MAX_WEIGHTED_MATCHING_ORDERING_METIS_PQ(); break;
        case 58: L.set_NORM_MAX_WEIGHTED_MATCHING_ORDERING_METIS_PQ(); break;
        case 59: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SYMMPQ(); break;
        case 60: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SYMMPQ(); break;
        case 61: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMC(); break;
        case 62: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMC(); break;
        case 63: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MC(); break;
        case 64: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MC(); break;
        case 65: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMCI(); break;
        case 66: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMCI(); break;
        case 67: L.set_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MCI(); break;
        case 68: L.set_NORM_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MCI(); break;
        #endif
        #ifdef ILUPLUSPLUS_USES_PARDISO
        case 69:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING(); break;
        case 70:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING(); break;
        case 71:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
        case 72:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_PQ(); break;
        case 73:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
        case 74:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER(); break;
        case 75:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
        case 76:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_MOVE_CORNER_IM(); break;
        case 77:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
        case 78:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR(); break;
        case 79:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
        case 80:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT_MOV_COR_IM(); break;
        case 81:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
        case 82:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR(); break;
        case 83:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
        case 84:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_WGT2_MOV_COR_IM(); break;
        case 85:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        case 86:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_DD_MOV_COR_IM(); break;
        case 87:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
        case 88:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SYM_PQ(); break;
        case 89:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
        case 90:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER(); break;
        case 91:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
        case 92:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SYMB_MOVE_CORNER_IM(); break;
        case 93:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
        case 94:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER(); break;
        case 95:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;
        case 96:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_SP_MOVE_CORNER_IM(); break;
        #endif
        #if defined(ILUPLUSPLUS_USES_METIS) && defined(ILUPLUSPLUS_USES_PARDISO)
        case 97:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER(); break;
        case 98:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER(); break;
        case 99:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER_IM(); break;
        case 100:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_MOVE_CORNER_IM(); break;
        case 101:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR(); break;
        case 102:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR(); break;
        case 103:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR_IM(); break;
        case 104:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_WGT_MOV_COR_IM(); break;
        case 105:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_DD_MOV_COR_IM(); break;
        case 106:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_DD_MOV_COR_IM(); break;
        case 107:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR(); break;
        case 108:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR(); break;
        case 109:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR_IM(); break;
        case 110:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SP_MOV_COR_IM(); break;
        case 111:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_METIS(); break;
        case 112:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_METIS(); break;
        case 113:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_METIS_PQ(); break;
        case 114:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_ORDERING_METIS_PQ(); break;
        case 115:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SYMMPQ(); break;
        case 116:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SYMMPQ(); break;
        case 117:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMC(); break;
        case 118:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMC(); break;
        case 119:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MC(); break;
        case 120:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MC(); break;
        case 121:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMCI(); break;
        case 122:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_SMCI(); break;
        case 123:  L.set_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MCI(); break;
        case 124:  L.set_NORM_PARDISO_MAX_WEIGHTED_MATCHING_METIS_ORDERING_W2MCI(); break;
        #endif  // end PARDISO & METIS available
        default:  std::cout<<"Unknown value for preprocessing type. Using default."<<std::endl; L.set_MAX_WEIGHTED_MATCHING_ORDERING_UNIT_DIAG(); break;
    }

    param.init(L,atoi(argv[2]),""); 
    param.set_MEM_FACTOR(atof(argv[3]));
    param.set_fill_in(atoi(argv[4]));	
    //param.set_MAX_FILLIN_IS_INF(true);	
    param.set_threshold(atof(argv[5]));
	param.set_BEGIN_TOTAL_PIV(false);



    // declare (sparse) matrices and (dense) vectors
    iluplusplus::matrix Acol;
    iluplusplus::vector X,X_exact,B;

    Acol.read_hb(argv[6]);
    B.read_hb(argv[6],1);

    bool have_rhs = (B.dim()>0);

    // make artificial RHS, if we do not have one.
    if (have_rhs){
        std::cout<<"  RHS read successfully."<<std::endl;
    } else {
        std::cout<<"  RHS not read successfully. Making artificial RHS."<<std::endl;
        X_exact.resize(Acol.rows(),1.0);
        // X_exact.set_all(1.0);
        // works too but requires copying:
        // B = Acol*X_exact;
        // better, especially for large problems:
        Acol.multiply(X_exact,B);
    }

    // set dropping parameter. set_threshold(tau) sets dropping threshold to 10^{-tau}

    // set first tolerance for terminating iterative method (also used as return variable for reduction of residual obtained, i.e. the relative residual)
    iluplusplus::Real rel_tol   = 8.0;  // actual tolerance is 10^{-rel_tol}
    // set second tolerance for terminating iterative method (also used as return variable for residual obtained, i.e. the absolute residual)
    iluplusplus::Real abs_tol   = 8.0;  // actual tolerance is 10^{-abs_tol}
    // variable for returning the absolute error (if exact solution is known -- otherwise nan is returned)
    iluplusplus::Real abs_error;
    // maximal number of iterations. The variable returns the number of iterations required for convergence
    iluplusplus::Integer max_iter = 600;

    std::string output_directory   = argv[7];
    std::string subdirectory_data  = (output_directory + argv[8]);
    std::string matrix_name        = argv[9];

    bool success;
    success = iluplusplus::solve_with_multilevel_preconditioner(Acol,B,X_exact,X,!have_rhs,rel_tol,abs_tol,max_iter,abs_error,output_directory,matrix_name,param,true,subdirectory_data);

    if(success) {
        std::cout<<"        Solved in "<<max_iter<<" iterations and achieved an absolute residual of 10^("<<-abs_tol;
        std::cout<<") and reduction of the initial residual by 10^("<<-rel_tol<<")"<<std::endl;
        //std::cout<<"The calculated solution is "<<std::endl<<x;
    } else {
        std::cout<<"Iteration failed."<<std::endl;
    }

	//------------------------------------------------------------------------
	//							ILUC
	//------------------------------------------------------------------------
	
	//----------------------iluplusplus::matrix
    iluplusplus::Integer min_iter = 1;
    // thes variables need to be reset, as they contain the results of the previous call of iluplusplus::solve_with_multilevel_preconditioner
    abs_tol = 8.0; rel_tol = 8.0; max_iter = 600;
    // declare preconditioner
	iluplusplus::Integer fill_in = param.get_fill_in();
	iluplusplus::Real threshold  = param.get_threshold();
	iluplusplus::matrix Arow;
	Arow.change_orientation_of_data(Acol);
    iluplusplus::iluc Pr1(Arow,fill_in,threshold);

    std::cout<<"                Preconditioner has a fill-in of "<<(iluplusplus::Real) Pr1.total_nnz() / (iluplusplus::Real) Arow.non_zeroes()<<std::endl;
    std::cout<<"                Preconditioner requires approximately "<< Pr1.memory()<<" bytes to store."<<std::endl;
    std::cout<<"                Preconditioner calculation required approximately "<< Pr1.memory_used_calculations()<<" bytes."<<std::endl;
    std::cout<<"                Approximately "<< Pr1.memory_allocated_calculations()<<" bytes were allocated."<<std::endl<<std::endl;


    // call solver
    // Note: BiCGstab returns rel_tol abs_tol achieved
	X.set_all(0.0);
    BiCGstab(Pr1,Arow,B,X,min_iter,max_iter,rel_tol,abs_tol);
	

    std::cout<<"        Solved in "<<max_iter<<" iterations and achieved an absolute residual of 10^(-"<<abs_tol;
    std::cout<<") and reduction of the initial residual by 10^(-"<<rel_tol<<")"<<std::endl;	

	//----------------------CSR
    int *ia = 0;             // make sure that this type is of iluplusplus::Integer
    int *ja = 0;             // make sure that this type is of iluplusplus::Integer
    Coeff_Field *a = 0;      // make sure that this type corresponds to the Coeff_Field of iluplusplus matrices and vectors 
	
   // setup matrix in sparse compressed row format

     ia  = new int[9];
          ia[0] = 0; ia[1] = 4; ia[2] = 7; ia[3] = 9; ia[4] = 11; ia[5] = 12; ia[6] = 15; ia[7] = 17; ia[8] = 20;

     ja = new int[20];
          ja[0]  = 0; ja[1]  = 2; ja[2]  = 5; ja[3]  = 6;
          ja[4]  = 1; ja[5]  = 2; ja[6]  = 4;
          ja[7]  = 2; ja[8]  = 7;
          ja[9]  = 3; ja[10] = 6;
          ja[11] = 1;
          ja[12] = 2; ja[13] = 5; ja[14] = 7;
          ja[15] = 1; ja[16] = 6;
          ja[17] = 2; ja[18] = 6; ja[19] = 7;

     a = new iluplusplus::Coeff_Field[20];
          a[0]  = (Coeff_Field) 7.0;  a[1]  =  (Coeff_Field) 1.0; a[2]  =  (Coeff_Field) 2.0; a[3]  =  (Coeff_Field) 7.0;
          a[4]  = (Coeff_Field) -4.0; a[5]  =  (Coeff_Field) 8.0; a[6]  =  (Coeff_Field) 2.0;
          a[7]  = (Coeff_Field) 1.0;  a[8]  =  (Coeff_Field) 5.0;
          a[9]  = (Coeff_Field) 7.0;  a[10] =  (Coeff_Field) 9.0;
          a[11] = (Coeff_Field) -4.0;
          a[12] = (Coeff_Field)  7.0; a[13] =  (Coeff_Field) 3.0; a[14] =  (Coeff_Field) 8.0;
          a[15] = (Coeff_Field)  1.0; a[16] =  (Coeff_Field) 11.0;
          a[17] = (Coeff_Field) -3.0; a[18] =  (Coeff_Field) 2.0; a[19] =  (Coeff_Field) 5.0;

    // Setup right hand side

    Coeff_Field *rhs = 0;
    rhs = new iluplusplus::Coeff_Field[8];
           rhs[0]  =  (Coeff_Field) 17.0; rhs[1]  =   (Coeff_Field) 6.0;  rhs[2]  =   (Coeff_Field) 6.0;  rhs[3]  =  (Coeff_Field) 16.0;
           rhs[4]  =  (Coeff_Field) -4.0; rhs[5]  =   (Coeff_Field) 18.0; rhs[6]  =   (Coeff_Field) 12.0; rhs[7]  =  (Coeff_Field) 4.0;
	
    iluplusplus::matrix A;
	iluplusplus::vector B2,B3;	
	
    int nrow = 8;
    int ncol = 8;
	int nrhs = 8; 
    int nnz = 20;
    iluplusplus::orientation_type orientation = iluplusplus::ROW;

	// make ILU++ matrix and vector type
	B2.interchange(rhs,nrhs);	
	B3 = B2;
	A.interchange(a,ja,ia,nrow,ncol,nnz,orientation);	
	
	iluplusplus::iluc Pr2(A,fill_in,threshold);	
	
    std::cout<<"                Preconditioner has a fill-in of "<<(iluplusplus::Real) Pr2.total_nnz() / (iluplusplus::Real) A.non_zeroes()<<std::endl;
    std::cout<<"                Preconditioner requires approximately "<< Pr2.memory()<<" bytes to store."<<std::endl;
    std::cout<<"                Preconditioner calculation required approximately "<< Pr2.memory_used_calculations()<<" bytes."<<std::endl;
    std::cout<<"                Approximately "<< Pr2.memory_allocated_calculations()<<" bytes were allocated."<<std::endl<<std::endl;

    // call solver
    // Note: BiCGstab returns rel_tol abs_tol achieved
	X.set_all(0.0);
    BiCGstab(Pr2,A,B2,X,min_iter,max_iter,rel_tol,abs_tol);

    std::cout<<"        Solved in "<<max_iter<<" iterations and achieved an absolute residual of 10^(-"<<abs_tol;
    std::cout<<") and reduction of the initial residual by 10^(-"<<rel_tol<<")"<<std::endl;	

	//------------------------------------------------------------------------
	//							ILUT
	//------------------------------------------------------------------------
	
	iluplusplus::ilut Pr3(A,0,threshold);	

    std::cout<<"                Preconditioner has a fill-in of "<<(iluplusplus::Real) Pr3.total_nnz() / (iluplusplus::Real) A.non_zeroes()<<std::endl;
    std::cout<<"                Preconditioner requires approximately "<< Pr3.memory()<<" bytes to store."<<std::endl;
    std::cout<<"                Preconditioner calculation required approximately "<< Pr3.memory_used_calculations()<<" bytes."<<std::endl;
    std::cout<<"                Approximately "<< Pr3.memory_allocated_calculations()<<" bytes were allocated."<<std::endl<<std::endl;

    // call solver
    // Note: BiCGstab returns rel_tol abs_tol achieved
	X.set_all(0.0);
    BiCGstab(Pr3,A,B3,X,min_iter,max_iter,rel_tol,abs_tol);
	
    std::cout<<"        Solved in "<<max_iter<<" iterations and achieved an absolute residual of 10^(-"<<abs_tol;
    std::cout<<") and reduction of the initial residual by 10^(-"<<rel_tol<<")"<<std::endl;	
	
    // interchange again	
	A.interchange(a,ja,ia,nrow,ncol,nnz,orientation);	
	B2.interchange(rhs,nrhs);	
	
    delete [] a;
    delete [] ia;
    delete [] ja;
    delete [] rhs;
	
    return 0;
  }
  catch(iluplusplus::iluplusplus_error ippe){
      std::cerr<<"main: ERROR: "<<ippe.error_message()<<". Returning."<<std::endl;
      return 1;
  }
  catch(std::bad_alloc){
      std::cerr<<"main: Error allocating memory. Returning."<<std::endl;
      return 1;
  }
}
